#ifndef XMMS_SYNTH_FILTER_H
#define XMMS_SYNTH_FILTER_H

#include "mpc_dec.h"

void  Synthese_Filter_opt ( short* OutData, float VV [V_MEM + 960], float YY [36] [32], int Channels );

#endif /* XMMS_SYNTH_FILTER_H */
