#ifndef XMMS_MPC_DEC_H
#define XMMS_MPC_DEC_H

#define   V_MEM     2304

extern float         ClipPrevFactor;
extern unsigned int  ClipPrev_activated;

extern unsigned int  StreamVersion;
extern unsigned int  MS_used;
extern unsigned int  FwdJumpInfo;
extern unsigned int  ActDecodePos;
extern unsigned int  OverallFrames;
extern unsigned int  DecodedFrames;

int   DECODE                ( char* buffer, int* FrameWasValid, float V [2] [V_MEM + 960], float Y [2] [36] [32] );
void  RESET_Synthesis       ( void );
void  RESET_Globals         ( void );
int   Read_Bitstream        ( int SV );
int   Read_Bitstream_Jumper ( int SV );

#endif /* XMMS_MPC_DEC_H */
