#ifndef XMMS_HUFFSV7_H
#define XMMS_HUFFSV7_H

#include "bitstream.h"

void  Huffman_SV7_Encoder ( void );
void  Huffman_SV7_Decoder ( void );

extern HuffmanTyp         HuffHdr  [10];
extern HuffmanTyp         HuffSCFI [ 4];
extern HuffmanTyp         HuffDSCF [16];
extern const HuffmanTyp*  HuffQ    [ 2] [ 8];

#endif /* XMMS_HUFFSV7_H */
