#ifndef XMMS_HUFFSV46_H
#define XMMS_HUFFSV46_H

#include "bitstream.h"

void  Huffman_SV6_Encoder ( void );
void  Huffman_SV6_Decoder ( void );

extern const HuffmanTyp*  SampleHuff    [18];
extern const HuffmanTyp*  Region        [32];

extern HuffmanTyp         SCFI_Bundle   [ 8];
extern HuffmanTyp         DSCF_Entropie [13];
extern HuffmanTyp         Region_A      [16];
extern HuffmanTyp         Region_B      [ 8];
extern HuffmanTyp         Region_C      [ 4];

#endif /* XMMS_HUFFSV46_H */
